#include <octave/oct.h>
#include <octave/parse.h>
#include <octave/f77-fcn.h>

/* 

An octave gateway routine to the FORTRAN code rkf45.f explained in
detail in

Computer Methods for Mathematical Computations
by George E. Forsythe, Michael A. Malcolm, and Cleve B. Moler.
on pages 129--147

And found at

http://cm.bell-labs.com/netlib/ode/

Written by:
-- 
John L. Weatherwax                2005-10-29

email: wax@alum.mit.edu

Please send comments and especially bug reports to the
above email address.

-----
*/ 

typedef F77_RET_T (*fort_fcn_ptr_dec) (const double&,double*, double*);

extern "C"
{
  F77_RET_T F77_FUNC(rkf45,RKF45)(fort_fcn_ptr_dec, const int&, double*, 
				  double&, double&, 
				  const double&, const double&, 
				  int&, double*, int* );
}

static octave_function *rkf45g_oct_fcn;
static int NEQN; 

static F77_RET_T
rkf45g_f_fcn(const double& t, double* x, double* xp)
{
  RowVector rv_x(NEQN); 
  for( int ii=0; ii < NEQN; ii++ ) rv_x(ii)=x[ii];

  octave_value_list args;
  args(0) = t; args(1) = rv_x; 
  int nargout = 1;

  octave_value_list retval = feval(rkf45g_oct_fcn, args, nargout);

  if (! error_state && retval.length () == 1)
    {
      Matrix m_xp = retval(0).matrix_value();
      for (int ii = 0; ii < NEQN; ii++){
	xp[ii] = m_xp(ii);
      }

      if (error_state)
	error ("failure evaluating user-supplied function");
    }
  else
    error ("failure evaluating user-supplied function");

  F77_RETURN(0);
}

// C++ octave API calls of the user's input function:
//
ColumnVector
rkf45_f_function(double t, const ColumnVector& x)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x;

  if (rkf45g_oct_fcn)
    {
      octave_value_list tmp = rkf45g_oct_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("rkf45g_oct_fcn");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).vector_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("rkf45g_oct_fcn");
	}
      else
	gripe_user_supplied_eval ("rkf45g_oct_fcn");
    }

  return retval;
}


DEFUN_DLD (rkf45g, args, nargout,
  "x_t = rkf45g(fcn, x_0, t): Integrates dx/dt = fcn from 0 to t")
{
  octave_value_list retval;

  if( args.length() != 3 ){ print_usage("rkf45g"); return retval; }
  if( nargout > 1 ) { print_usage("rkf45g"); return retval; }

  // Extract the user input:
  //
  rkf45g_oct_fcn = extract_function(args(0), "rkf45g", "__rkf45g_oct_fcn__",
	     "function F = __rkf45g_oct_fcn__ (t, x) ", "; endfunction");
  if(error_state){
    error( "rkf45g: expecting a function for the first argument" ); 
  }
  ColumnVector x_0( args(1).vector_value() );
  if(error_state){
    error("rkf45g: expecting a vector for the second argument" ); 
  }
  NEQN = x_0.length(); 
  double TFINAL = args(2).double_value();
  if(error_state){
    error("rkf45g: expecting a scalar for the third argument" ); 
  }

  // FORTRAN variables:
  double T=0.0; 
  double RELERR = 1.e-9;
  double ABSERR = 0.0;
  //  double TFINAL = 10.0; 
  double TPRINT = 0.5;
  int IFLAG = +1; 

  // Create the work/iwork vectors: 
  int NWORK = 3+6*NEQN;
  OCTAVE_LOCAL_BUFFER( double, WORKP, NWORK ); 
  int NIWORK = 5;
  OCTAVE_LOCAL_BUFFER( int, IWORKP, NIWORK );
  OCTAVE_LOCAL_BUFFER( double, XP, NEQN ); 
  for( int ii=0; ii < NEQN; ii++ ) XP[ii]=x_0(ii);

  // Create octave output matrix: 
  int nouts = 1 + ( (int) ( (TFINAL-T)/TPRINT ) + 1 );
  Matrix y_out(NEQN,nouts);
  for( int ii=0; ii < NEQN; ii++ ) y_out(ii,0)=x_0(ii);

  // Initialization call to RKF45:
  double TOUT=T;
  F77_FUNC(rkf45,RKF45)(rkf45g_f_fcn,NEQN,XP,T,TOUT,\
			RELERR,ABSERR,IFLAG,WORKP,IWORKP);

  for( int iit=1; iit < nouts; iit++ ){ // Perform each integration step:
    TOUT = ( (T+TPRINT)<=TFINAL ? (T+TPRINT) : TFINAL );
    F77_FUNC(rkf45,RKF45)(rkf45g_f_fcn,NEQN,XP,T,TOUT,\
			  RELERR,ABSERR,IFLAG,WORKP,IWORKP);

    ColumnVector x_out(NEQN); 
    for( int ii=0; ii < NEQN; ii++ ){
      y_out(ii,iit) = XP[ii];
    }
  }

  retval.append(y_out); 
  return retval;
}
